/* (c) Copyright 2004-2005, Cadence Design Systems, Inc.  All rights reserved. 

This file is part of the OA Gear distribution.  See the COPYING file in
the top level OA Gear directory for copyright and licensing information. */

/*
Author: Aaron Hurst <ahurst@eecs.berkeley.edu>

ChangeLog:
2005-08-18: ChangeLog started
*/

#include <iostream>
#include "oaDesignDB.h"
#include "oagFunc.h"
#include "oagFuncManager.h"
#include "oagUtilOption.h"
#include "oagUtilOptionParser.h"
#include "oagMapperTable.h"

using namespace std;


// *****************************************************************************
// findOpenLib()
//
/// \brief Find an already open, opens from disk, or creates a library.
///
// *****************************************************************************
oa::oaLib *
findOpenLib(const oa::oaScalarName libName, const char *libString) {
    oa::oaLib *lib = NULL;

    if ((lib = oa::oaLib::find(libName)))
        ;
    else if (!oa::oaLib::exists(libString)) {
        lib = oa::oaLib::create(libName, libString);
        oa::oaLibDefList *defList = oa::oaLibDefList::get("lib.defs", 'a');
        oa::oaLibDef::create(defList, libName, libString);
        defList->save();
    } else {
        lib = oa::oaLib::open(libName, libString); 
    }

    return lib;
}


// *****************************************************************************
// findOpenDesign()
//
/// \brief Find an already open, opens from disk, or creates a design.
///
/// Opens designs from disk for read.
//
// *****************************************************************************
oa::oaDesign *
findOpenDesign(oa::oaScalarName libName, oa::oaScalarName viewName,
               const char *cellString) {
    const oa::oaNativeNS nativeNS;
    const oa::oaScalarName cellName(nativeNS, cellString);
    oa::oaDesign *design = NULL;

    if ((design = oa::oaDesign::find(libName, cellName, viewName))) {
        ;
    } else {
        // open for read only
        try {
            return oa::oaDesign::open(libName, cellName, viewName, 'r');
        } catch (oa::oaException &e) {}
    }
    return design;
}


// *****************************************************************************
// main()
//
// *****************************************************************************
int main(int argc, const char **argv) {

    std::cout << "*****************************************************************************" << std::endl;
    std::cout << "Tool:   tableMap        Package:    oagMapper   version:  01/2007" << std::endl;
    std::cout << "*****************************************************************************" << std::endl;
    std::cout << std::endl;

    // Define the options of interest

    oagUtil::OptionParser options("Map a design to set of arbitrary library cells.");

    oagUtil::Option *objectiveOpt = options.add("obj", "Mapping objective", 
            true, "area|delay");
    oagUtil::Option *designLibOpt = options.add("design_lib", "Library "
            "name of design to be mapped", true, "lib");
    oagUtil::Option *designViewOpt = options.add("design_view", "View "
            "name of the design to be mapped (default: netlist)", false, 
            "view");
    oagUtil::Option *cellOpt = options.add("cell", "Cell name of "
            "the design to be mapped", true, "cell");
    oagUtil::Option *libraryLibOpt = options.add("library_lib", "Library name "
            "of technology library ", true, "library");
    oagUtil::Option *libraryViewOpt = options.add("library_view", "View name "
            "with tech library functional info (default: netlist)", false, "view");
    oagUtil::Option *instanceViewOpt = options.add("instance_view", "View name "
            "for mapped library instances (default: netlist)", false, "view");
    oagUtil::Option *seqCellOpt = options.add("seq_cell", "Sequential cell "
            "used for mapping", true, "cell");
    oagUtil::Option *clockNetOpt = options.add("clock_net", "Manually specify global clock net name",
             false, "net");
    oagUtil::Option *clockTriggerOpt = options.add("clock_trig", "Manually specify global "
            "clock net trigger type (default: pos)", false, "pos|neg");
    oagUtil::Option *resetNetOpt = options.add("reset_net", "Manually specify global reset net name",
             false, "net");
    oagUtil::Option *resetTriggerOpt = options.add("reset_trig", "Manually specify global reset net "
            "trigger type (default: pos)", false, "pos|neg");
    oagUtil::Option *flattenOpt = options.add("flatten", "Flatten the cell "
            "before mapping", false);
    oagUtil::Option *cutsPerNodeOpt = options.add("cuts_per_node", "Maximum number "
            "of cuts to generate per node (default=200)", false, "int");
    oagUtil::Option *saveOpt = options.add("save", "Save the design", false);

    if (!options.parse(argc, argv)) {
        cerr << options.getMessage();
        exit(0);
    }

    // Get the names supplied, or fill in the default values.

    const char *libraryLibString = libraryLibOpt->getValue();
    const char *designLibString = designLibOpt->getValue();
    const char *libraryViewString 
        = libraryViewOpt->isGiven() ? libraryViewOpt->getValue() : "netlist";
    const char *instanceViewString 
        = instanceViewOpt->isGiven() ? instanceViewOpt->getValue() : "netlist";
    const char *designViewString 
        = designViewOpt->isGiven() ? designViewOpt->getValue() : "netlist";
    const char *cellString = cellOpt->getValue();

    try {

        // Initialize OpenAccess

        oa::oaDesignInit();
        oa::oaLibDefList::openLibs();
        oagFunc::initialize();

        const oa::oaNativeNS nativeNS;
        const oa::oaVerilogNS verilogNS;
        const oa::oaScalarName libraryLibName(nativeNS, libraryLibString);
        const oa::oaScalarName libraryViewName(nativeNS, libraryViewString);
        const oa::oaScalarName instanceViewName(nativeNS, instanceViewString);
        const oa::oaScalarName designLibName(nativeNS, designLibString);
        const oa::oaScalarName designViewName(nativeNS, designViewString);
        
        // Find, create, or load the libraries.

        oa::oaLib *library = findOpenLib(libraryLibName, libraryLibString);
        assert(library);
        library->getAccess(oa::oacReadLibAccess);
            
        library = findOpenLib(designLibName, designLibString);
        assert(library);
        library->getAccess(oa::oacWriteLibAccess);
    
        // Find or open the design to map

        const oa::oaScalarName cellName(nativeNS, cellString);
        oa::oaDesign *targetDesign;
        if ((targetDesign = oa::oaDesign::find(designLibName, cellName, designViewName)))
            ;
        else if (saveOpt->isGiven()) {
            // open for read and write
            targetDesign = oa::oaDesign::open(designLibName, cellName, designViewName, 'a');
        } else {
            // open for read only
            targetDesign = oa::oaDesign::open(designLibName, cellName, designViewName, 'r');
        }    

        // Initialize mapper

        oa::oaDesign *seqDesign = findOpenDesign(libraryLibName, 
                libraryViewName, seqCellOpt->getValue());
        if (!seqDesign) {
            std::cerr << "ERROR: Could not open sequential cell" << std::endl;
        }

        int cutsPerNode = 200;
        if (cutsPerNodeOpt->isGiven()) {
          cutsPerNode = atoi(cutsPerNodeOpt->getValue());
        }

        oagMapper::Table mapper(cutsPerNode);
        mapper.mapUtils.setSeq(seqDesign);
        mapper.useAlternateView(instanceViewName);
        
        // Find or open the library cells

        cout << "Combinataional logic cells: ";
        oa::oaLib *libraryLib = oa::oaLib::find(libraryLibName);
        assert(libraryLib);
        libraryLib->getAccess(oa::oacReadLibAccess);
        oa::oaCell *libraryCell;
        oa::oaIter<oa::oaCell> cellIter(libraryLib->getCells());
        while((libraryCell = cellIter.getNext())) {
          oa::oaScalarName cellName;
          libraryCell->getName(cellName);
          oa::oaString cellString;
          cellName.get(cellString);

          oa::oaDesign *design = findOpenDesign(libraryLibName, 
                                                libraryViewName, 
                                                cellString);        

          if (design) {
            // ignore cells with sequential nodes
            oagFunc::Manager *manager = oagFunc::Manager::get(design);
            if (manager && manager->getLocalSeqNodeCount() > 0) continue;
            
            cout << cellString << " ";
            mapper.addLibraryGate(design);
          }
        }
        cout << endl;

        // Flatten

        if (flattenOpt->isGiven()) {
            cout << "Flattening \'" << cellString << "\' ... " << endl;
            oagFunc::collapseAllInstancesAllLevels(targetDesign->getTopModule());
        }

        // Manually identify clocks/resets

        if (clockNetOpt->isGiven()) {
            oa::oaScalarName clockName(verilogNS, clockNetOpt->getValue());
            oa::oaModNet *clockNet = oa::oaModNet::find(targetDesign->getTopModule(), clockName);
            if (!clockNet) {
                cerr << "ERROR: Could not find clock net: " << clockNetOpt->getValue() << endl;
                exit(0);
            }
            if (clockNet->getNumBits() != 1) {
                cerr << "ERROR: Clock net not a single bit net: " << clockNetOpt->getValue() << endl;
                exit(0);                
            }
            oagMapper::Utils::TriggerType clockTrigger = oagMapper::Utils::POSEDGE;
            if (clockTriggerOpt->isGiven()) {
                if (!strcmp(clockTriggerOpt->getValue(), "pos")) {
                    clockTrigger = oagMapper::Utils::POSEDGE;
                } else if (!strcmp(clockTriggerOpt->getValue(), "neg")) {
                    clockTrigger = oagMapper::Utils::NEGEDGE;
                } else {
                    cerr << "ERROR: Unknown trigger type.  Use \"pos\" or \"neg\"" << endl;
                    exit(0);
                }
            }
            mapper.mapUtils.addGlobalClock(oagFunc::toBitNet(clockNet), clockTrigger);
        }
        if (resetNetOpt->isGiven()) {
            oa::oaScalarName resetName(verilogNS, resetNetOpt->getValue());
            oa::oaModNet *resetNet = oa::oaModNet::find(targetDesign->getTopModule(), resetName);
            if (!resetNet) {
                cerr << "ERROR: Could not find reset net: " << resetNetOpt->getValue() << endl;
                exit(0);
            }
            if (resetNet->getNumBits() != 1) {
                cerr << "ERROR: Reset net not a single bit net: " << resetNetOpt->getValue() << endl;
                exit(0);
            }
            oagMapper::Utils::TriggerType resetTrigger = oagMapper::Utils::POSEDGE;
            if (resetTriggerOpt->isGiven()) {
                if (!strcmp(resetTriggerOpt->getValue(), "pos")) {
                    resetTrigger = oagMapper::Utils::POSEDGE;
                } else if (!strcmp(resetTriggerOpt->getValue(), "neg")) {
                    resetTrigger = oagMapper::Utils::NEGEDGE;
                } else {
                    cerr << "ERROR: Unknown trigger type.  Use \"pos\" or \"neg\"" << endl;
                    exit(0);
                }
            }
            mapper.mapUtils.addGlobalReset(oagFunc::toBitNet(resetNet), resetTrigger);
        }
        
        // Map!

        if (!strcmp(objectiveOpt->getValue(), "area")) {
          mapper.techmapArea(targetDesign->getTopModule());
        } else if (!strcmp(objectiveOpt->getValue(), "delay")) {
          mapper.techmapDelay(targetDesign->getTopModule());
        } else {
          cerr << "ERROR: Can't understand mapping objective.  Try \"area\" or \"delay\"" << endl;
          exit(0);
        }
        mapper.mapUtils.printGateUsage(targetDesign->getTopModule());

        // Remove dangling nets

        mapper.mapUtils.removeDanglingNets(targetDesign->getTopModule());

        // Create a pointer to technology database in library

        oa::oaLibDMData *libData = oa::oaLibDMData::open(designLibName, 'a');
        assert(libData);
        oa::oaStringProp::create(libData, "techLibName", libraryLibString);
        libData->save();
        libData->purge();

        // Save 

        if (saveOpt->isGiven()) {
            cout << "Saving \'" << cellString << "\' ... " << endl;
            targetDesign->save();
        }

        // Close

        oa::oaDesign *openDesign;
        oa::oaIter<oa::oaDesign> designIter(oa::oaDesign::getOpenDesigns());
        while((openDesign = designIter.getNext())) {
            oa::oaString designNameString;
            openDesign->getTopModule()->getName(nativeNS, designNameString);
            openDesign->purge();
        }
        std::cout << std::endl;

    } catch(oa::oaException &e) {
        cerr << "ERROR: " << e.getMsg() << endl;
        return 1;
    }

    return 0;
}
